#include <girepository.h>
#include <ctype.h>

#import <ObjFW/ObjFW.h>

#define NS "Gtk"

OFString *camel_case(OFString *name)
{
	OFMutableString *ret = [OFMutableString string];
	OFArray *parts = [name componentsSeparatedByString: @"_"];
	OFEnumerator *enumerator = [parts objectEnumerator];
	OFString *part;

	part = [enumerator nextObject];
	[ret appendString: part];
	while (part = [enumerator nextObject]) {
		[ret appendString: [[part substringWithRange: of_range(0, 1)] uppercaseString]];
		[ret appendString: [part substringWithRange: of_range(1, [part length] - 1)]];
	}

	[ret makeImmutable];
	return ret;
}

void print_method(GIFunctionInfo *meth, OFFile *file)
{
	GITypeInfo *ret_type = g_callable_info_get_return_type(meth);
	GITypeTag ret_typeTag = g_type_info_get_tag(ret_type);

	GIFunctionInfoFlags flags = g_function_info_get_flags(meth);
	if (flags & GI_FUNCTION_IS_METHOD)
		[file writeString: @"-"];
	else
		[file writeString: @"+"];

	[file writeFormat: @" (%s)", g_type_tag_to_string(ret_typeTag)];
	[file writeString: camel_case([OFString stringWithUTF8String: g_base_info_get_name(meth)])];

	gint arg_count = g_callable_info_get_n_args(meth);
	if (arg_count)
		[file writeString: @"With"];
	for (gint j = 0; j < arg_count; j++) {
		GIArgInfo *arg = g_callable_info_get_arg(meth, j);
		GITypeInfo *type = g_arg_info_get_type(arg);
		GITypeTag typeTag = g_type_info_get_tag(type);
		const gchar *name = g_base_info_get_name(arg);
		/* GIDirection dir = g_arg_info_get_direction(arg);
		if (dir == GI_DIRECTION_IN) printf("in ");
		if (dir == GI_DIRECTION_OUT) printf("out ");
		if (dir == GI_DIRECTION_INOUT) printf("inout "); */
		[file writeString: camel_case([OFString stringWithFormat: @"%c%s: ", (j>0) ? *name : toupper(*name), name + 1])];
		[file writeFormat: @"(%s%s)", g_type_tag_to_string(typeTag), g_type_info_is_pointer(type) ? "*" : ""];
		[file writeFormat: @"%s", name];
		if (j < arg_count - 1)
			[file writeString: @"\n"];
	}
	[file writeString: @";\n"];
}

void print_methods(GIObjectInfo *info, OFFile *file)
{
	gint count = g_object_info_get_n_methods(info);
	for (gint i = 0; i < count; i++) {
		GIFunctionInfo *meth = g_object_info_get_method(info, i);
		print_method(meth, file);
	}
}

int main(int argc, char *argv[])
{
	GIRepository *repository;
	GITypelib *typelib;

	g_type_init();

	repository = g_irepository_get_default();
	typelib = g_irepository_require(repository, NS, NULL, 0, NULL);

	if (!typelib) {
		[of_stderr writeLine: @"Namespace could not be loaded"];
		return -1;
	}

	gint count = g_irepository_get_n_infos(repository, NS);

	for (gint i = 0; i < count; i++) {
		GIBaseInfo *info = g_irepository_get_info(repository, NS, i);
		if (g_base_info_get_type(info) == GI_INFO_TYPE_OBJECT) {
			OFString *filename = [OFString stringWithFormat: @"header/%s%s.h", NS, g_base_info_get_name(info)];
			OFFile *header = [OFFile fileWithPath: filename
							 mode: @"w"];
			[header writeFormat: @"@interface %s%s : OFObject\n", NS, g_base_info_get_name(info)];
			print_methods((GIObjectInfo*) info, header);
			[header writeLine: @"@end"];
			[header close];
		}
	}

	return 0;
}
